@@ -27,7 +27,7 @@ if prefix <> '@' + 'prefix@':
         libdir = interpolate (libdir) % vars()
 
 # run from textrace-source dir.
-
+exit_value = 0
 simplify_p = 0
 verbose_p = 0
 pfa_p = 0
@@ -36,8 +36,16 @@ afm_p = 0
 dos_kpath_p = 0
 ttf_p = 0
 keep_trying_p = 0
-# should be installed.
-autotrace_bin = 'autotrace'
+backend_options = ''
+
+fontforge_cmd = 'fontforge'
+
+# You can take this higher, but then rounding errors will have
+# nasty side effects.
+# Used as reciprocal grid size
+potrace_scale = 1.0
+round_to_int = 1
+
 magnification = 1000.0
 program_name = 'mftrace'
 temp_dir = os.path.join (os.getcwd (), program_name + '.dir' )
@@ -48,14 +56,23 @@ program_version='@VERSION@'
 origdir= os.getcwd ()
 
 coding_dict = {
-	'Extended TeX Font Encoding - Latin' : 'tex256.enc',
-	'TeX text': 'tex256.enc',
 
+	# from TeTeX
+	'TeX typewriter text' : '09fbbfac.enc', # cmtt10
+	'TeX math symbols':'10037936.enc ', # cmbsy
+	'ASCII caps and digits':'1b6d048e', # cminch
+	'TeX math italic': 'aae443f0.enc ',  # cmmi10
+	'TeX extended ASCII':'d9b29452.enc',
+	'TeX text': 'f7b6d320.enc',
+	'TeX text without f-ligatures' : '0ef0afca.enc',
 	
-	'feta braces': 'feta-braces0.enc',
-	'feta number': 'feta-nummer10.enc',
-	'feta music': 'feta20.enc',
-	'parmesan music' : 'parmesan20.enc',
+	'Extended TeX Font Encoding - Latin' : 'tex256.enc',
+
+	# LilyPond.
+	'fetaBraces': 'feta-braces-a.enc',
+	'fetaNumber': 'feta-nummer10.enc',
+	'fetaMusic': 'feta20.enc',
+	'parmesanMusic' : 'parmesan20.enc',
 	}
 
 
@@ -106,8 +123,6 @@ def warning (s):
 	errorport.write (_ ("warning: ") + s)
 		
 def error (s):
-
-
 	'''Report the error S.  Exit by raising an exception. Please
 	do not abuse by trying to catch this error. If you do not want
 	a stack trace, write to the output directly.
@@ -261,23 +276,29 @@ Example:
 option_definitions = [
 	('', 'h', 'help', _ ("This help")),
 	('', 'k', 'keep', _ ("Keep all output in directory %s.dir") % program_name),
+	('', '', 'keep-all', _ ("Really keep all output (implies --keep)")),
 	('MAG', '', 'magnification', _("Set magnification for MF to MAG (default: 1000).")),
 	('', 'V', 'verbose', _ ("Verbose")),
 	('', 'v', 'version', _ ("Print version number")),
 	('', 'a', 'pfa', _ ("Generate PFA file (default)")),
 	('','', 'afm',  _("Generate AFM file (implies --simplify)")),
 	('', 'b', 'pfb', _ ("Generate PFB file")),
-	('', '', 'simplify', _("Simplify using pfaedit")),
+	('', '', 'simplify', _("Simplify using fontforge")),
 	('FILE', '', 'gffile', _("Use gf FILE instead of running Metafont")),
 	('DIR', 'I', 'include', _("Add to path for searching files")),
 	('LIST','', 'glyphs', _('Process only these glyphs. LIST is comma separated')),
 	('FILE', '', 'tfmfile' , _("Use FILE for the TFM file")),
 	('FILE', 'o', 'output-base', _("Set output file name")), 
 	('ENC', 'e', 'encoding', _("Use encoding file ENC")),
-	('', 't', 'truetype', _("Generate TrueType file (requires pfaedit).")),
-	('', '', 'keep-trying', _("Don't stop if autotrace fails")),
+	('', 't', 'truetype', _("Generate TrueType file (requires fontforge).")),
+	('', '', 'keep-trying', _("Don't stop if tracing fails")),
 	('', 'w', 'warranty', _ ("show warranty and copyright")),
 	('', '', 'dos-kpath', _ ("try to use Miktex kpsewhich")),
+	('', '', 'potrace', _ ("Use potrace")),
+	('', '', 'autotrace', _ ("Use autotrace")),
+	('', '', 'noround', _ ("Don't round coordinates of control points to integer values (use with --grid)")),
+	('GRID', '', 'grid', _ ("Set reciprocal grid size in em units"))
+	
 	]
 
 
@@ -310,73 +331,61 @@ def find_file (nm):
 # TRACING.
 ################################################################
 
-def autotrace_command (fi, fn, opts):
-	opts = " " + opts + " --output-format=eps --input-format=pbm "
-	cmd = autotrace_bin + opts + " --output-file=char.eps -filter-iterations %d %s  " % (fi,fn)
-	return cmd
+def autotrace_command (fn, opts):
+	opts = " " + opts + " --background-color=FFFFFF --output-format=eps --input-format=pbm "
+	return trace_binary + opts + backend_options + " --output-file=char.eps  %s  " % fn
 
-def run_autotrace  (fi,fn,opts):
-	stat = system (autotrace_command (fi,fn,opts), 1)
-	return stat
 
-def do_autotrace_best_fi (fn, opts):
+def potrace_command (fn, opts):
+	return trace_binary + opts \
+	      + ' -u %d ' % potrace_scale \
+	      + backend_options + "  -q -c --eps --output=char.eps %s " % (fn) 
 
-	""" Run autotrace, and find the
-	best filter-iterations value.
+trace_command = None
+trace_binary = '' 
+path_to_type1_ops = None
+
+
+
+def trace_one (pbmfile, id):
 	"""
-	fi = 8
-	while fi >= 0:
-		if run_autotrace (fi, fn, opts) ==0:
-			return 0
-
-		fi  = fi -1  
-
-	return 1
-
-def blank_pbm (filename):
+	Run tracer, do error handling 
 	"""
-	Kill the contents of a PBM: write 0xFF to the PBM.
-	"""
-	f = open (filename)
-	l = f.readline ()
-	length = len (f.read())
 
-	f.close ()
-	
-	open (filename, 'w').write ('%s\n%s' % (l,'\377' * length))
+	status = system (trace_command (pbmfile, ''), 1)
 
-def do_autotrace_one (pbmfile, id):
-	"""
-	Run autotrace, first with -background-color, then without.
-	"""
-	
-	status = run_autotrace (9, pbmfile , '-background-color FFFFFF')
 	if status == 2:
 		sys.stderr.write ("\nUser interrupt. Exiting\n")
 		sys.exit(2)
+
+	if status == 0 and keep_temp_dir_p:
+		shutil.copy2 (pbmfile, '%s.pbm' % id)
+		shutil.copy2 ('char.eps', '%s.eps' % id)
+		
 	if status <> 0:
-		error_file = os.path.join (origdir, 'autotrace-bug-%s.pbm' % id)
+		error_file = os.path.join (origdir, 'trace-bug-%s.pbm' % id)
 		shutil.copy2 (pbmfile, error_file)
-		msg = """Autotrace failed on bitmap. Bitmap left in `%s\'
+		msg = """Trace failed on bitmap. Bitmap left in `%s\'
 Failed command was:
 
 	%s
 	
-Please submit a bugreport to autotrace development.""" % (error_file,
-							  autotrace_command (9, error_file, '-background-color FFFFFF'))
+Please submit a bugreport to %s development.""" % (error_file, 
+						   trace_command (error_file, ''),
+						   trace_binary)
 
 		if keep_trying_p:
 			warning (msg)
 			sys.stderr.write ("\nContinuing trace...\n")
+			exit_value = 1
 		else:
 			msg = msg + '\nRun mftrace with --keep-trying to produce a font anyway\n'
 			error (msg)
 	else:
 		return 1
 
-	status = do_autotrace_best_fi (pbmfile, '-background-color FFFFFF')
 	if status <> 0:
-		warning ("Failing even  -filter-iterations=0. Skipping character.\n")
+		warning ("Failed, skipping character.\n")
 		return 0
 	else:
 		return 1
@@ -392,7 +401,7 @@ def make_pbm (filename, outname, char_number):
 	status = system (command, ignore_error = 1)
 	
 	return  (status == 0)
-
+ 
 def read_encoding (file):
 	sys.stderr.write(_("Using encoding file: `%s'\n") % file)
 	
@@ -403,9 +412,9 @@ def read_encoding (file):
 	if not m:
 		raise 'Encoding file invalid.'
 	
-	name = m.group(1)	
-	cod =m.group( 2)
-	cod = re.sub('[ /]+', ' ',cod)
+	name = m.group (1)	
+	cod = m.group (2)
+	cod = re.sub ('[ /]+', ' ',cod)
 	cods = string.split (cod)
 
 	return (name, cods)
@@ -444,7 +453,6 @@ def autotrace_path_to_type1_ops (at_file, bitmap_metrics, tfm_wid):
 	commands = []
 	
 
-
 	while ls[0] <> '*U\n':
 		l = ls[0]
 		ls = ls[1:]
@@ -455,15 +463,16 @@ def autotrace_path_to_type1_ops (at_file, bitmap_metrics, tfm_wid):
 			continue
 		cmd= toks[-1]
 		args = map (lambda m, s=inv_scale : s * string.atof(m), toks[:-1])
-		args = zip_to_pairs (map (round, args))
+		if round_to_int:
+			args = zip_to_pairs (map (round, args))
+		else :
+			args = zip_to_pairs (args)
 		commands.append ((cmd,args))
 
-
 	expand = {
 		'l': 'rlineto',
 		'm': 'rmoveto',
 		'c': 'rrcurveto',
-		'sbw' : 'sbw',
 		'f': 'closepath' ,
 		}
 
@@ -474,7 +483,7 @@ def autotrace_path_to_type1_ops (at_file, bitmap_metrics, tfm_wid):
 	t1_outline =  '  %d %d hsbw\n' % (- off_x, tfm_wid)
 	bbox =  (10000,10000,-10000,-10000)
 
-	for (c,args) in commands[0:]:
+	for (c,args) in commands:
 
 		na = []
 		for a in args:
@@ -487,7 +496,10 @@ def autotrace_path_to_type1_ops (at_file, bitmap_metrics, tfm_wid):
 
 		a = na
 		c = expand[c]
-		a = map (lambda x: '%d' % int (x),  unzip_pairs (a))
+		if round_to_int:
+			a = map (lambda x: '%d' % int (round (x)),  unzip_pairs (a))
+		else :
+			a = map (lambda x: '%d %d div' % (int (round (x * potrace_scale/inv_scale)), int( round (potrace_scale/inv_scale))),  unzip_pairs (a))
 
 		t1_outline = t1_outline + '  %s %s\n' % (string.join (a),c)
 
@@ -495,6 +507,98 @@ def autotrace_path_to_type1_ops (at_file, bitmap_metrics, tfm_wid):
 	t1_outline = '{\n %s } |- \n' % t1_outline
 	
 	return (bbox, t1_outline)
+
+def potrace_path_to_type1_ops (at_file, bitmap_metrics, tfm_wid):
+	inv_scale = 1000.0/magnification
+	
+	(size_y, size_x, off_x, off_y)= map(lambda m, s=inv_scale : m * s, bitmap_metrics)
+	ls = open (at_file).readlines ()
+	bbox =  (10000,10000,-10000,-10000)
+
+	while ls and ls[0] <> '0 setgray\n':
+		ls = ls[1:]
+
+	if ls == []:
+		return (bbox, '')
+	ls = ls[1:]
+	commands = []
+	
+	while ls and ls[0] <> 'grestore\n':
+		l = ls[0]
+		ls = ls[1:]
+
+		if l == 'fill\n':
+			continue
+		
+		toks = string.split (l)
+
+		if len(toks) < 1:
+			continue
+		cmd= toks[-1]
+		args = map (lambda m, s=inv_scale : s * string.atof (m), toks[:-1])
+		args = zip_to_pairs (args)
+		commands.append ((cmd,args))
+
+
+
+	# t1asm seems to fuck up when using sbw. Oh well. 
+	t1_outline =  '  %d %d hsbw\n' % (- off_x, tfm_wid)
+	bbox =  (10000,10000,-10000,-10000)
+
+	# Type1 fonts have relative coordinates (doubly relative for
+	# rrcurveto), so must convert moveto and rcurveto.
+
+	z = (0.0,  size_y - off_y - 1.0)
+	nc = []
+	for (c, args) in commands:
+		args = map (lambda x: (x[0] * (1.0  / potrace_scale), x[1] * (1.0/potrace_scale)), args)
+		
+		if c == 'moveto':
+			args = [(args[0][0]-z[0], args[0][1] - z[1])]
+			
+		zs = []
+		for a in args:
+			lz= (z[0] + a[0], z[1] + a[1])
+			bbox = update_bbox_with_point (bbox, lz)
+			zs.append (lz)
+
+		if round_to_int:
+			last_discr_z = (int (round (z[0])), int (round (z[1])))
+		else :
+			last_discr_z = (z[0], z[1])
+		args = []
+		for a in zs:
+			if round_to_int:
+				a = (int (round (a[0])), int (round (a[1])))
+			else :
+				a = (a[0], a[1])
+			args.append ( (a[0] - last_discr_z[0],
+				       a[1] - last_discr_z[1]))
+
+			last_discr_z = a
+
+		if zs:
+			z = zs[-1]
+		c = { 'rcurveto': 'rrcurveto',
+		      'moveto': 'rmoveto',
+		      'closepath': 'closepath',
+		      'rlineto' : 'rlineto'}[c]
+
+		if c == 'rmoveto':
+			t1_outline += ' closepath '
+	
+		if round_to_int:
+			args = map (lambda x: '%d' % int (round (x)),  unzip_pairs (args))
+		else :
+			args = map (lambda x: '%d %d div' % (int (round (x * potrace_scale/inv_scale)), int( round (potrace_scale/inv_scale))),  unzip_pairs (args))
+
+		t1_outline = t1_outline + '  %s %s\n' % (string.join (args),c)
+
+	t1_outline = t1_outline + ' endchar '
+	t1_outline = '{\n %s } |- \n' % t1_outline
+	
+	return (bbox, t1_outline)
+
 	
 def read_gf_dims (name, c):
 	str = popen ('gf2pbm -n %d -s %s' % (c, name)).read ()
@@ -503,8 +607,7 @@ def read_gf_dims (name, c):
 	return tuple (map (string.atoi ,m.groups ()))
 		       
 
-	
-def autotrace_font (fontname, gf_file, metric, glyphs, encoding, magnification):
+def trace_font (fontname, gf_file, metric, glyphs, encoding, magnification):
 	t1os = []
 	font_bbox =  (10000,10000,-10000,-10000)
 
@@ -530,7 +633,7 @@ def autotrace_font (fontname, gf_file, metric, glyphs, encoding, magnification):
 			sys.stderr.flush()
 
 		# this wants the id, not the filename.
-		success = do_autotrace_one ("char.pbm", '%s-%d' % (gf_fontname, a))
+		success = trace_one ("char.pbm", '%s-%d' % (gf_fontname, a))
 		if not success :
 			sys.stderr.write ("(skipping character)]")
 			sys.stderr.flush ()			
@@ -541,7 +644,7 @@ def autotrace_font (fontname, gf_file, metric, glyphs, encoding, magnification):
 			sys.stderr.flush()
 		metric_width = metric.get_char (a).width
 		tw = int (round (metric_width / metric.design_size * 1000))
-		(bbox, t1o)  = autotrace_path_to_type1_ops ("char.eps",
+		(bbox, t1o)  = path_to_type1_ops ("char.eps",
 						    (h, w, xo, yo),
 						    tw)
 
@@ -690,13 +793,24 @@ def update_bbox_with_bbox (bb, dims):
 		
 	return (llx,lly,urx,ury) 
 
+
 def check_pfaedit_scripting ():
-	stat = system ("pfaedit -usage > pfv 2>&1", ignore_error = 1)
+	global fontforge_cmd
+	stat = system ("fontforge -usage > pfv 2>&1 > /dev/null", ignore_error = 1)
 	if stat <> 0:
-		warning ("Command `pfaedit -usage' failed. Cannot simplify or convert to TTF.")
+		stat = system ("pfaedit -usage > pfv 2>&1", ignore_error = 1)
+		if stat == 0:
+			fontforge_cmd = 'pfaedit'
+	else:
+		fontforge_cmd = 'fontforge'
+
+	if stat <> 0:
+		warning ("Command `fontforge -usage' failed. Cannot simplify or convert to TTF.")
 		return 0
-		
-	if re.search ("-script", open ('pfv').read()) == None:
+
+	
+	if fontforge_cmd == 'pfaedit' and \
+	       re.search ("-script", open ('pfv').read()) == None:
 		warning ("pfaedit does not support -script. Install 020215 or later.\nCannot simplify or convert to TTF.\n")
 		return 0
 	return 1
@@ -714,16 +828,17 @@ def cleanup_font (file):
 	
 	progress (_ ("Simplifying font... "))
 	
-	open ('simplify.pe', 'w').write ('''#!/usr/bin/env pfaedit
+	open ('simplify.pe', 'w').write ('''#!/usr/bin/env %s
 Open ($1);
 MergeKern($2);
 SelectAll ();
 Simplify ();
 AutoHint ();
+RoundToInt();
 Generate ("%s");
 Quit (0);
-''' % file)
-	system ("pfaedit -script simplify.pe %s %s" % (file, tfmfile))
+''' % (fontforge_cmd, file))
+	system ("%s -script simplify.pe %s %s" % (fontforge_cmd, file, tfmfile))
 	progress ('\n')
 
 def make_ttf (fontname):
@@ -737,17 +852,18 @@ def make_ttf (fontname):
 	# not used?
 	shutil.copy2 (fontname + '.pfx', "before-pfaedit.pfx")
 	
-	open ('to-ttf.pe', 'w').write ('''#!/usr/bin/env pfaedit
+	open ('to-ttf.pe', 'w').write ('''#!/usr/bin/env %s
 Open ($1);
 MergeKern($2);
 SelectAll ();
 Simplify ();
 AutoHint ();
+RoundToInt();
 Generate ("%s");
 Quit (0);
-''' % (filename + '.ttf'))
+''' % (fontforge_cmd, filename + '.ttf'))
 	
-	system ("pfaedit -script to-ttf.pe %s %s" % ((filename + '.pfx'), tfmfile))
+	system ("%s -script to-ttf.pe %s %s" % (fontforge_cmd, (filename + '.pfx'), tfmfile))
 	
 
 def getenv (var, default):
@@ -760,7 +876,7 @@ def getenv (var, default):
 def gen_pixel_font (filename, metric, magnification):
 	"""
 	Generate a GF file  for FILENAME, such that `magnification'*mfscale
-	(default 1000 * 1.0) points fit on the designsize.
+	(default 1000 * 1.0) pixels fit on the designsize.
 	"""
 	base_dpi = 600
 
@@ -780,7 +896,7 @@ def gen_pixel_font (filename, metric, magnification):
 
 		progress (_ ("Running Metafont..."))
 
-		cmdstr = r"mf '\mode:=ljfour; mag:=%f; nonstopmode; input %s'" %  (mag,filename)
+		cmdstr = r"mf '\mode:=ljfour; mag:=%f; nonstopmode; input %s'" %  (mag, filename)
 		if not verbose_p:
 			cmdstr = cmdstr +  ' 1>/dev/null 2>/dev/null'
 		st = system (cmdstr, ignore_error = 1)
@@ -875,9 +991,9 @@ This should be tailored for each metafont font set.
 
 	fontinfo['ItalicAngle'] = 0
 	if re.search ('[Ii]talic', shape) or re.search ('[Oo]blique', shape):
-		a = 14
+		a = -14
 		if re.search ("Sans", family):
-			a = 12
+			a = -12
 			
 		fontinfo ["ItalicAngle"] = a
 		
@@ -962,10 +1078,36 @@ for (o,a) in options:
 	elif o == '--afm':
 		afm_p = 1
 		simplify_p = 1
-		
+	elif o == '--potrace':
+		trace_binary = 'potrace'
+	elif o == '--autotrace' :
+		trace_binary = 'autotrace'
+	elif o == '--noround' :
+		round_to_int = 0
+	elif o == '--grid' :
+		potrace_scale = round (string.atof(a))
 	else:
 		raise 'Ugh -- forgot to implement option %s. :)' % o
 
+
+	
+backend_options = getenv('MFTRACE_BACKEND_OPTIONS','')
+
+stat = os.system ('potrace --version > /dev/null') 
+if trace_binary <> 'autotrace' and stat == 0:
+	trace_binary = 'potrace'
+	trace_command = potrace_command
+	path_to_type1_ops = potrace_path_to_type1_ops
+
+stat = os.system ('autotrace --version > /dev/null') 
+if trace_binary <> 'potrace' and stat == 0:
+	trace_binary = 'autotrace'
+	trace_command = autotrace_command
+	path_to_type1_ops = autotrace_path_to_type1_ops
+
+if not trace_binary:
+	error (_("No tracing program found. Exit."))
+
 identify (sys.stderr)
 if not pfa_p and not pfb_p and not ttf_p:
 	pfa_p = 1
@@ -1039,7 +1181,7 @@ for filename in files:
 		gf_fontname = find_file (gf_fontname)
 
 	# the heart of the program:
-	autotrace_font (basename, gf_fontname, metric, glyph_range, encoding, magnification)
+	trace_font (basename, gf_fontname, metric, glyph_range, encoding, magnification)
 
 	if pfa_p:
 		if simplify_p:
@@ -1062,3 +1204,5 @@ for filename in files:
 
 	os.chdir (origdir)
 	cleanup_temp ()
+
+sys.exit (exit_value)
